@@ -50,6 +50,8 @@ static PyObject * decode_bytestring(CBORDecoderObject *, uint8_t);
 static PyObject * decode_string(CBORDecoderObject *, uint8_t);
 static PyObject * CBORDecoder_decode_datetime_string(CBORDecoderObject *);
 static PyObject * CBORDecoder_decode_epoch_datetime(CBORDecoderObject *);
+static PyObject * CBORDecoder_decode_epoch_date(CBORDecoderObject *);
+static PyObject * CBORDecoder_decode_date_string(CBORDecoderObject *);
 static PyObject * CBORDecoder_decode_fraction(CBORDecoderObject *);
 static PyObject * CBORDecoder_decode_bigfloat(CBORDecoderObject *);
 static PyObject * CBORDecoder_decode_rational(CBORDecoderObject *);
@@ -347,30 +349,69 @@ _CBORDecoder_get_immutable(CBORDecoderObject *self, void *closure)
 // Utility functions /////////////////////////////////////////////////////////
 
 static int
-fp_read(CBORDecoderObject *self, char *buf, const Py_ssize_t size)
-{
-    PyObject *obj, *size_obj;
-    char *data;
-    int ret = -1;
+raise_from(const PyObject *new_exc_type, const char *message) {
+    // This requires the error indicator to be set
+    PyObject *cause;
+#if PY_VERSION_HEX >= 0x030c0000
+    cause = PyErr_GetRaisedException();
+#else
+    PyObject *exc_type, *exc_traceback;
+    PyErr_Fetch(&exc_type, &cause, &exc_traceback);
+    PyErr_NormalizeException(&exc_type, &cause, &exc_traceback);
+    Py_XDECREF(exc_type);
+    Py_XDECREF(exc_traceback);
+#endif
 
+    PyObject *msg_obj = PyUnicode_FromString(message);
+    if (message) {
+        PyObject *new_exception = PyObject_CallFunctionObjArgs(
+            new_exc_type, msg_obj, NULL);
+        if (new_exception) {
+            PyException_SetCause(new_exception, cause);
+            PyErr_SetObject(new_exc_type, new_exception);
+        }
+        Py_DECREF(msg_obj);
+    }
+}
+
+static PyObject *
+fp_read_object(CBORDecoderObject *self, const Py_ssize_t size)
+{
+    PyObject *ret = NULL;
+    PyObject *obj, *size_obj;
     size_obj = PyLong_FromSsize_t(size);
     if (size_obj) {
         obj = PyObject_CallFunctionObjArgs(self->read, size_obj, NULL);
+        Py_DECREF(size_obj);
         if (obj) {
             assert(PyBytes_CheckExact(obj));
             if (PyBytes_GET_SIZE(obj) == (Py_ssize_t) size) {
-                data = PyBytes_AS_STRING(obj);
-                memcpy(buf, data, size);
-                ret = 0;
+                ret = obj;
             } else {
+                Py_DECREF(obj);
                 PyErr_Format(
                     _CBOR2_CBORDecodeEOF,
                     "premature end of stream (expected to read %zd bytes, "
                     "got %zd instead)", size, PyBytes_GET_SIZE(obj));
             }
-            Py_DECREF(obj);
         }
-        Py_DECREF(size_obj);
+    }
+    return ret;
+}
+
+
+static int
+fp_read(CBORDecoderObject *self, char *buf, const Py_ssize_t size)
+{
+    int ret = -1;
+    PyObject *obj = fp_read_object(self, size);
+    if (obj) {
+        char *data = PyBytes_AS_STRING(obj);
+        if (data) {
+            memcpy(buf, data, size);
+            ret = 0;
+        }
+        Py_DECREF(obj);
     }
     return ret;
 }
@@ -536,17 +577,12 @@ decode_negint(CBORDecoderObject *self, uint8_t subtype)
 
 
 static PyObject *
-decode_definite_bytestring(CBORDecoderObject *self, Py_ssize_t length)
+decode_definite_short_bytestring(CBORDecoderObject *self, Py_ssize_t length)
 {
-    PyObject *ret = NULL;
-
-    ret = PyBytes_FromStringAndSize(NULL, length);
+    PyObject *ret = fp_read_object(self, length);
     if (!ret)
         return NULL;
-    if (fp_read(self, PyBytes_AS_STRING(ret), length) == -1) {
-        Py_DECREF(ret);
-        return NULL;
-    }
+
     if (string_namespace_add(self, ret, length) == -1) {
         Py_DECREF(ret);
         return NULL;
@@ -555,6 +591,56 @@ decode_definite_bytestring(CBORDecoderObject *self, Py_ssize_t length)
 }
 
 
+static PyObject *
+decode_definite_long_bytestring(CBORDecoderObject *self, Py_ssize_t length)
+{
+    PyObject *buffer = NULL;
+    Py_ssize_t left = length;
+    while (left) {
+        Py_ssize_t chunk_length = length <= 65536 ? length : 65536;
+        PyObject *chunk = fp_read_object(self, chunk_length);
+        if (!chunk) {
+            break;
+        }
+
+        if (!PyBytes_CheckExact(chunk)) {
+            Py_DECREF(chunk);
+            break;
+        }
+
+        if (buffer) {
+            PyObject *new_buffer = PyByteArray_Concat(buffer, chunk);
+            Py_DECREF(chunk);
+            if (!new_buffer)
+                break;
+
+            if (new_buffer != buffer) {
+                Py_DECREF(buffer);
+                buffer = new_buffer;
+            }
+        } else {
+            buffer = PyByteArray_FromObject(chunk);
+            Py_DECREF(chunk);
+            if (!buffer)
+                break;
+        }
+        left -= chunk_length;
+    }
+
+    PyObject *ret = NULL;
+    if (buffer) {
+        ret = PyBytes_FromObject(buffer);
+        Py_DECREF(buffer);
+
+        if (ret && string_namespace_add(self, ret, length) == -1) {
+            Py_DECREF(ret);
+            ret = NULL;
+        }
+    }
+    return ret;
+}
+
+
 static PyObject *
 decode_indefinite_bytestrings(CBORDecoderObject *self)
 {
@@ -613,9 +699,14 @@ decode_bytestring(CBORDecoderObject *self, uint8_t subtype)
     }
     if (indefinite)
         ret = decode_indefinite_bytestrings(self);
+    else if (length <= 65536)
+        ret = decode_definite_short_bytestring(self, (Py_ssize_t)length);
     else
-        ret = decode_definite_bytestring(self, (Py_ssize_t)length);
-    set_shareable(self, ret);
+        ret = decode_definite_long_bytestring(self, (Py_ssize_t)length);
+
+    if (ret)
+        set_shareable(self, ret);
+
     return ret;
 }
 
@@ -635,21 +726,16 @@ decode_bytestring(CBORDecoderObject *self, uint8_t subtype)
 
 
 static PyObject *
-decode_definite_string(CBORDecoderObject *self, Py_ssize_t length)
+decode_definite_short_string(CBORDecoderObject *self, Py_ssize_t length)
 {
-    PyObject *ret = NULL;
-    char *buf;
+    PyObject *bytes_obj = fp_read_object(self, length);
+    if (!bytes_obj)
+        return NULL;
 
-    buf = PyMem_Malloc(length);
-    if (!buf)
-        return PyErr_NoMemory();
-
-    if (fp_read(self, buf, length) == 0)
-        ret = PyUnicode_DecodeUTF8(
-                buf, length, PyBytes_AS_STRING(self->str_errors));
-    PyMem_Free(buf);
-
-    if (string_namespace_add(self, ret, length) == -1) {
+    const char *bytes = PyBytes_AS_STRING(bytes_obj);
+    PyObject *ret = PyUnicode_FromStringAndSize(bytes, length);
+    Py_DECREF(bytes_obj);
+    if (ret && string_namespace_add(self, ret, length) == -1) {
         Py_DECREF(ret);
         return NULL;
     }
@@ -657,6 +743,104 @@ decode_definite_string(CBORDecoderObject *self, Py_ssize_t length)
 }
 
 
+static PyObject *
+decode_definite_long_string(CBORDecoderObject *self, Py_ssize_t length)
+{
+    PyObject *ret = NULL, *chunk = NULL, *string = NULL;
+    Py_ssize_t left = length;
+    Py_ssize_t consumed;
+    Py_ssize_t buffer_size = 0;  // how many bytes are allocated for the buffer
+    Py_ssize_t buffer_length = 0;  // how many bytes are actually stored in the buffer
+    char *buffer = NULL;
+    while (left) {
+        // Read up to 65536 bytes of data from the stream
+        Py_ssize_t chunk_length = 65536 - buffer_size;
+        if (left < chunk_length)
+            chunk_length = left;
+
+        PyObject *chunk = fp_read_object(self, chunk_length);
+        left -= chunk_length;
+        if (!chunk)
+            goto error;
+
+        // Get the internal buffer of the bytes object
+        char *bytes_buffer = PyBytes_AsString(chunk);
+        if (!bytes_buffer)
+            goto error;
+
+        char *source_buffer;
+        if (buffer) {
+            // Grow the buffer to accommodate the previous data plus the new chunk
+            if (buffer_length + chunk_length > buffer_size) {
+                buffer_size = buffer_length + chunk_length;
+                char *new_buffer = PyMem_Realloc(buffer, buffer_size);
+                if (!new_buffer)
+                    goto error;
+
+                buffer = new_buffer;
+            }
+
+            // Concatenate the chunk into the buffer
+            memcpy(buffer + buffer_length, bytes_buffer, chunk_length);
+            buffer_length += chunk_length;
+
+            source_buffer = buffer;
+            chunk_length = buffer_length;
+        } else {
+            // Use the chunk's internal buffer directly to decode as many characters as possible
+            source_buffer = bytes_buffer;
+        }
+
+        string = PyUnicode_DecodeUTF8Stateful(source_buffer, chunk_length, NULL, &consumed);
+        if (!string)
+            goto error;
+
+        if (ret) {
+            // Concatenate the result to the existing result
+            PyObject *joined = PyUnicode_Concat(ret, string);
+            if (!joined)
+                goto error;
+
+            Py_DECREF(string);
+            string = NULL;
+            ret = joined;
+        } else {
+            // Set the result to the decoded string
+            ret = string;
+        }
+
+        Py_ssize_t unconsumed = chunk_length - consumed;
+        if (consumed != chunk_length) {
+            if (buffer) {
+                // Move the unconsumed bytes to the start of the buffer
+                memmove(buffer, buffer + consumed, unconsumed);
+            } else {
+                // Create a new buffer
+                buffer = PyMem_Malloc(unconsumed);
+                if (!buffer)
+                    goto error;
+
+                memcpy(buffer, bytes_buffer + consumed, unconsumed);
+            }
+            buffer_length = unconsumed;
+        }
+    }
+
+    if (ret && string_namespace_add(self, ret, length) == -1)
+        goto error;
+
+    return ret;
+error:
+    Py_XDECREF(ret);
+    Py_XDECREF(chunk);
+    Py_XDECREF(string);
+    if (buffer)
+        PyMem_Free(buffer);
+
+    return NULL;
+}
+
+
 static PyObject *
 decode_indefinite_strings(CBORDecoderObject *self)
 {
@@ -714,8 +898,14 @@ decode_string(CBORDecoderObject *self, uint8_t subtype)
     }
     if (indefinite)
         ret = decode_indefinite_strings(self);
+    else if (length <= 65536)
+        ret = decode_definite_short_string(self, (Py_ssize_t)length);
     else
-        ret = decode_definite_string(self, (Py_ssize_t)length);
+        ret = decode_definite_long_string(self, (Py_ssize_t)length);
+
+    if (!ret && PyErr_GivenExceptionMatches(PyErr_Occurred(), PyExc_UnicodeDecodeError))
+        raise_from(_CBOR2_CBORDecodeValueError, "error decoding unicode string");
+
     set_shareable(self, ret);
     return ret;
 }
@@ -936,11 +1126,12 @@ decode_map(CBORDecoderObject *self, uint8_t subtype)
     }
     if (ret && self->object_hook != Py_None) {
         map = PyObject_CallFunctionObjArgs(self->object_hook, self, ret, NULL);
-        if (map) {
-            set_shareable(self, map);
-            Py_DECREF(ret);
-            ret = map;
-        }
+        if (!map)
+            return NULL;
+
+        set_shareable(self, map);
+        Py_DECREF(ret);
+        ret = map;
     }
     return ret;
 }
@@ -970,10 +1161,12 @@ decode_semantic(CBORDecoderObject *self, uint8_t subtype)
             case 35:    ret = CBORDecoder_decode_regexp(self);          break;
             case 36:    ret = CBORDecoder_decode_mime(self);            break;
             case 37:    ret = CBORDecoder_decode_uuid(self);            break;
+            case 100:   ret = CBORDecoder_decode_epoch_date(self);      break;
             case 256:   ret = CBORDecoder_decode_stringref_ns(self);    break;
             case 258:   ret = CBORDecoder_decode_set(self);             break;
             case 260:   ret = CBORDecoder_decode_ipaddress(self);       break;
             case 261:   ret = CBORDecoder_decode_ipnetwork(self);       break;
+            case 1004:  ret = CBORDecoder_decode_date_string(self);     break;
             case 55799: ret = CBORDecoder_decode_self_describe_cbor(self);
                 break;
 
@@ -1005,7 +1198,7 @@ decode_semantic(CBORDecoderObject *self, uint8_t subtype)
 
 
 static PyObject *
-parse_datestr(CBORDecoderObject *self, PyObject *str)
+parse_datetimestr(CBORDecoderObject *self, PyObject *str)
 {
     const char* buf;
     char *p;
@@ -1078,6 +1271,31 @@ parse_datestr(CBORDecoderObject *self, PyObject *str)
     return ret;
 }
 
+static PyObject *
+parse_datestr(CBORDecoderObject *self, PyObject *str)
+{
+    const char* buf;
+    Py_ssize_t size;
+    PyObject *ret = NULL;
+    unsigned long int Y, m, d;
+
+    buf = PyUnicode_AsUTF8AndSize(str, &size);
+    if (
+            size < 10 || buf[4] != '-' || buf[7] != '-')
+    {
+        PyErr_Format(
+            _CBOR2_CBORDecodeValueError, "invalid date string %R", str);
+        return NULL;
+    }
+    if (buf) {
+        Y = strtoul(buf, NULL, 10);
+        m = strtoul(buf + 5, NULL, 10);
+        d = strtoul(buf + 8, NULL, 10);
+        ret = PyDate_FromDate(Y, m, d);
+    }
+    return ret;
+}
+
 
 // CBORDecoder.decode_datetime_string(self)
 static PyObject *
@@ -1086,6 +1304,63 @@ CBORDecoder_decode_datetime_string(CBORDecoderObject *self)
     // semantic type 0
     PyObject *match, *str, *ret = NULL;
 
+    if (!_CBOR2_datetimestr_re && _CBOR2_init_re_compile() == -1)
+        return NULL;
+    str = decode(self, DECODE_NORMAL);
+    if (str) {
+        if (PyUnicode_Check(str)) {
+            match = PyObject_CallMethodObjArgs(
+                    _CBOR2_datetimestr_re, _CBOR2_str_match, str, NULL);
+            if (match) {
+                if (match != Py_None)
+                    ret = parse_datetimestr(self, str);
+                else
+                    PyErr_Format(
+                        _CBOR2_CBORDecodeValueError,
+                        "invalid datetime string: %R", str);
+                Py_DECREF(match);
+            }
+        } else
+            PyErr_Format(
+                _CBOR2_CBORDecodeValueError, "invalid datetime value: %R", str);
+        Py_DECREF(str);
+    }
+    set_shareable(self, ret);
+    return ret;
+}
+
+// CBORDecoder.decode_epoch_date(self)
+static PyObject *
+CBORDecoder_decode_epoch_date(CBORDecoderObject *self)
+{
+    // semantic type 100
+    PyObject *num, *tuple, *ret = NULL;
+
+    num = decode(self, DECODE_NORMAL);
+    if (num) {
+        if (PyNumber_Check(num)) {
+            tuple = PyTuple_Pack(1, PyNumber_Multiply(num, PyLong_FromLong(24 * 60 * 60)));
+            if (tuple) {
+                ret = PyDate_FromTimestamp(tuple);
+                Py_DECREF(tuple);
+            }
+        } else {
+            PyErr_Format(
+                _CBOR2_CBORDecodeValueError, "invalid timestamp value %R", num);
+        }
+        Py_DECREF(num);
+    }
+    set_shareable(self, ret);
+    return ret;
+}
+
+// CBORDecoder.decode_date_string(self)
+static PyObject *
+CBORDecoder_decode_date_string(CBORDecoderObject *self)
+{
+    // semantic type 0
+    PyObject *match, *str, *ret = NULL;
+
     if (!_CBOR2_datestr_re && _CBOR2_init_re_compile() == -1)
         return NULL;
     str = decode(self, DECODE_NORMAL);
@@ -1099,12 +1374,12 @@ CBORDecoder_decode_datetime_string(CBORDecoderObject *self)
                 else
                     PyErr_Format(
                         _CBOR2_CBORDecodeValueError,
-                        "invalid datetime string: %R", str);
+                        "invalid date string: %R", str);
                 Py_DECREF(match);
             }
         } else
             PyErr_Format(
-                _CBOR2_CBORDecodeValueError, "invalid datetime value: %R", str);
+                _CBOR2_CBORDecodeValueError, "invalid date value: %R", str);
         Py_DECREF(str);
     }
     set_shareable(self, ret);
@@ -1128,6 +1403,12 @@ CBORDecoder_decode_epoch_datetime(CBORDecoderObject *self)
             if (tuple) {
                 ret = PyDateTime_FromTimestamp(tuple);
                 Py_DECREF(tuple);
+                if (!ret && (
+                    PyErr_GivenExceptionMatches(PyErr_Occurred(), PyExc_OverflowError)
+                    || PyErr_GivenExceptionMatches(PyErr_Occurred(), PyExc_OSError)
+                    || PyErr_GivenExceptionMatches(PyErr_Occurred(), PyExc_ValueError)
+                ))
+                    raise_from(_CBOR2_CBORDecodeValueError, "error decoding datetime from epoch");
             }
         } else {
             PyErr_Format(
@@ -1367,16 +1648,22 @@ CBORDecoder_decode_rational(CBORDecoderObject *self)
     // NOTE: see semantic type 4
     tuple = decode(self, DECODE_IMMUTABLE | DECODE_UNSHARED);
     if (tuple) {
-        if (PyTuple_CheckExact(tuple) && PyTuple_GET_SIZE(tuple) == 2) {
-            ret = PyObject_CallFunctionObjArgs(
-                    _CBOR2_Fraction,
-                    PyTuple_GET_ITEM(tuple, 0),
-                    PyTuple_GET_ITEM(tuple, 1),
-                    NULL);
+        if (PyTuple_CheckExact(tuple)) {
+            ret = PyObject_Call(_CBOR2_Fraction, tuple, NULL);
+            set_shareable(self, ret);
+            if (!ret && (
+                PyErr_GivenExceptionMatches(PyErr_Occurred(), PyExc_TypeError)
+                || PyErr_GivenExceptionMatches(PyErr_Occurred(), PyExc_ZeroDivisionError)
+            ))
+                raise_from(_CBOR2_CBORDecodeValueError, "error decoding rational");
+        } else {
+            PyErr_SetString(
+                _CBOR2_CBORDecodeValueError,
+                "error decoding rational: input value was not a tuple"
+            );
         }
         Py_DECREF(tuple);
     }
-    set_shareable(self, ret);
     return ret;
 }
 
@@ -1394,6 +1681,8 @@ CBORDecoder_decode_regexp(CBORDecoderObject *self)
     if (pattern) {
         ret = PyObject_CallFunctionObjArgs(_CBOR2_re_compile, pattern, NULL);
         Py_DECREF(pattern);
+        if (!ret && PyErr_GivenExceptionMatches(PyErr_Occurred(), _CBOR2_re_error))
+            raise_from(_CBOR2_CBORDecodeValueError, "error decoding regular expression");
     }
     set_shareable(self, ret);
     return ret;
@@ -1416,6 +1705,8 @@ CBORDecoder_decode_mime(CBORDecoderObject *self)
             ret = PyObject_CallMethodObjArgs(parser,
                     _CBOR2_str_parsestr, value, NULL);
             Py_DECREF(parser);
+            if (!ret && PyErr_GivenExceptionMatches(PyErr_Occurred(), PyExc_TypeError))
+                raise_from(_CBOR2_CBORDecodeValueError, "error decoding MIME message");
         }
         Py_DECREF(value);
     }
@@ -1437,6 +1728,11 @@ CBORDecoder_decode_uuid(CBORDecoderObject *self)
     if (bytes) {
         ret = PyObject_CallFunctionObjArgs(_CBOR2_UUID, Py_None, bytes, NULL);
         Py_DECREF(bytes);
+        if (!ret && (
+            PyErr_GivenExceptionMatches(PyErr_Occurred(), PyExc_TypeError)
+            || PyErr_GivenExceptionMatches(PyErr_Occurred(), PyExc_ValueError)
+        ))
+            raise_from(_CBOR2_CBORDecodeValueError, "error decoding UUID value");
     }
     set_shareable(self, ret);
     return ret;
