@@ -1,14 +1,19 @@
+from __future__ import annotations
+
 import math
+import platform
 import re
 import struct
 import sys
 from binascii import unhexlify
-from datetime import datetime, timedelta, timezone
+from datetime import date, datetime, timedelta, timezone
 from decimal import Decimal
 from email.message import Message
 from fractions import Fraction
 from io import BytesIO
 from ipaddress import ip_address, ip_network
+from pathlib import Path
+from typing import Type, cast
 from uuid import UUID
 
 import pytest
@@ -226,6 +231,7 @@ def test_binary(impl, payload, expected):
         ("62225c", '"\\'),
         ("62c3bc", "\u00fc"),
         ("63e6b0b4", "\u6c34"),
+        pytest.param("7a00010001" + "61" * 65535 + "c3b6", "a" * 65535 + "ö", id="split_unicode"),
     ],
 )
 def test_string(impl, payload, expected):
@@ -233,6 +239,26 @@ def test_string(impl, payload, expected):
     assert decoded == expected
 
 
+@pytest.mark.parametrize(
+    "payload",
+    [
+        pytest.param("6198", id="short"),
+        pytest.param("7a00010000" + "61" * 65535 + "c3", id="long"),
+        pytest.param("7f6198ff", id="indefinite"),
+    ],
+)
+def test_string_invalid_utf8(impl, payload: str) -> None:
+    with pytest.raises(impl.CBORDecodeValueError, match="error decoding unicode string") as exc:
+        impl.loads(unhexlify(payload))
+
+    assert isinstance(exc.value.__cause__, UnicodeDecodeError)
+
+
+def test_string_oversized(impl) -> None:
+    with pytest.raises(impl.CBORDecodeEOF, match="premature end of stream"):
+        (impl.loads(unhexlify("aeaeaeaeaeaeaeaeae0108c29843d90100d8249f0000aeaeffc26ca799")),)
+
+
 @pytest.mark.parametrize(
     "payload, expected",
     [
@@ -341,6 +367,28 @@ def test_simple_val_as_key(impl):
 #
 
 
+@pytest.mark.parametrize(
+    "payload, expected",
+    [
+        (
+            "d903ec6a323031332d30332d3231",
+            date(2013, 3, 21),
+        ),
+        (
+            "d8641945e8",
+            date(2018, 12, 31),
+        ),
+    ],
+    ids=[
+        "date/string",
+        "date/timestamp",
+    ],
+)
+def test_date(impl, payload, expected):
+    decoded = impl.loads(unhexlify(payload))
+    assert decoded == expected
+
+
 @pytest.mark.parametrize(
     "payload, expected",
     [
@@ -425,6 +473,28 @@ def test_bad_datetime(impl):
     assert str(excinfo.value) == "invalid datetime string: '0000-123-01'"
 
 
+def test_datetime_overflow(impl):
+    with pytest.raises(impl.CBORDecodeError) as excinfo:
+        impl.loads(unhexlify("c11b9b9b9b0000000000"))
+
+    assert isinstance(excinfo.value.__cause__, OverflowError)
+
+
+def test_datetime_value_too_large(impl):
+    with pytest.raises(impl.CBORDecodeError) as excinfo:
+        impl.loads(unhexlify("c11b1616161616161616161616161616"))
+
+    assert isinstance(excinfo.value.__cause__, OSError)
+
+
+def test_datetime_date_out_of_range(impl):
+    with pytest.raises(impl.CBORDecodeError) as excinfo:
+        impl.loads(unhexlify("a6c11b00002401001b000000000000ff00"))
+
+    cause_exc_class = OSError if platform.system() == "Windows" else ValueError
+    assert isinstance(excinfo.value.__cause__, cause_exc_class)
+
+
 def test_datetime_timezone(impl):
     decoded = impl.loads(b"\xc0\x78\x192018-08-02T07:00:59+00:30")
     assert decoded == datetime(2018, 8, 2, 7, 0, 59, tzinfo=timezone(timedelta(minutes=30)))
@@ -467,12 +537,35 @@ def test_rational(impl):
     assert decoded == Fraction(2, 5)
 
 
+def test_rational_invalid_iterable(impl):
+    with pytest.raises(
+        impl.CBORDecodeValueError, match="error decoding rational: input value was not a tuple"
+    ):
+        impl.loads(unhexlify("d81e01"))
+
+
+def test_rational_zero_denominator(impl):
+    with pytest.raises(impl.CBORDecodeValueError, match="error decoding rational") as exc:
+        impl.loads(unhexlify("d81e820100"))
+
+    assert isinstance(exc.value.__cause__, ZeroDivisionError)
+
+
 def test_regex(impl):
     decoded = impl.loads(unhexlify("d8236d68656c6c6f2028776f726c6429"))
     expr = re.compile("hello (world)")
     assert decoded == expr
 
 
+def test_regex_unbalanced_parentheses(impl):
+    with pytest.raises(
+        impl.CBORDecodeValueError, match="error decoding regular expression"
+    ) as exc:
+        impl.loads(unhexlify("d8236c68656c6c6f2028776f726c64"))
+
+    assert isinstance(exc.value.__cause__, re.error)
+
+
 def test_mime(impl):
     decoded = impl.loads(
         unhexlify(
@@ -486,11 +579,32 @@ def test_mime(impl):
     assert decoded.get_payload() == "Hello =A4uro"
 
 
+def test_mime_invalid_type(impl):
+    with pytest.raises(impl.CBORDecodeValueError, match="error decoding MIME message") as exc:
+        impl.loads(unhexlify("d82401"))
+
+    assert isinstance(exc.value.__cause__, TypeError)
+
+
 def test_uuid(impl):
     decoded = impl.loads(unhexlify("d825505eaffac8b51e480581277fdcc7842faf"))
     assert decoded == UUID(hex="5eaffac8b51e480581277fdcc7842faf")
 
 
+def test_uuid_invalid_length(impl):
+    with pytest.raises(impl.CBORDecodeValueError, match="error decoding UUID value") as exc:
+        impl.loads(unhexlify("d8254f5eaffac8b51e480581277fdcc7842f"))
+
+    assert isinstance(exc.value.__cause__, ValueError)
+
+
+def test_uuid_invalid_type(impl):
+    with pytest.raises(impl.CBORDecodeValueError, match="error decoding UUID value") as exc:
+        impl.loads(unhexlify("d82501"))
+
+    assert isinstance(exc.value.__cause__, TypeError)
+
+
 @pytest.mark.parametrize(
     "payload, expected",
     [
@@ -690,6 +804,15 @@ def test_object_hook(impl):
     assert decoded.state == {"a": 3, "b": 5}
 
 
+def test_object_hook_exception(impl):
+    def object_hook(decoder, data):
+        raise RuntimeError("foo")
+
+    payload = unhexlify("A2616103616205")
+    with pytest.raises(RuntimeError, match="foo"):
+        impl.loads(payload, object_hook=object_hook)
+
+
 def test_load_from_file(impl, tmpdir):
     path = tmpdir.join("testdata.cbor")
     path.write_binary(b"\x82\x01\x0a")
@@ -809,3 +932,25 @@ def test_decimal_payload_unpacking(impl, data, expected):
     with pytest.raises(impl.CBORDecodeValueError) as exc_info:
         impl.loads(unhexlify(data))
     assert exc_info.value.args[0] == f"Incorrect tag {expected} payload"
+
+
+@pytest.mark.parametrize(
+    "payload",
+    [
+        pytest.param(
+            unhexlify("5b7fffffffffffff00"),
+            id="bytestring",
+        ),
+        pytest.param(
+            unhexlify("7b7fffffffffffff00"),
+            id="unicode",
+        ),
+    ],
+)
+def test_oversized_read(impl, payload: bytes, tmp_path: Path) -> None:
+    CBORDecodeEOF = cast(Type[Exception], getattr(impl, "CBORDecodeEOF"))
+    with pytest.raises(CBORDecodeEOF, match="premature end of stream"):
+        dummy_path = tmp_path / "testdata"
+        dummy_path.write_bytes(payload)
+        with dummy_path.open("rb") as f:
+            impl.load(f)
